/*******************************************************************************
* Copyright (c) 1993, 1995 by Charles Kooperberg.                              *
* All rights reserved. This function is part of the Logspline Density          *
* Estimation Program. It was written by Charles Kooperberg at the University   *
* of Washington and the University of California at Berkeley between 1988 and  *
* 1991. It is described in the paper: Kooperberg, Charles and Stone, Charles J.*
* `Logspline density estimation for censored data', Journal of Computational   *
* and Graphical Statistics, 1 (1992) 301-328.                                  *
* You are free to use this program, for non-commercial purposes only,          *
* under two conditions:                                                        *
* (1) This note is not to be removed                                           *
* (2) Publications using logspline computations should reference the           *
*  publication  mentioned above.                                               *
* For questions, please email clk@stat.washington.edu                          *
*                       Charles Kooperberg, November 30, 1995                  *
*******************************************************************************/
/* These 2 routines comoute the informationmatrix (info) the score
   vector (score) and C(theta) (ctheta).  The routine consists
   of 2 parts. The middle1 routine, that does the numerical integrations
   in the middle, and tails that does the exact computations for
   the tails.                                                                 */

/* computes the numerical integrals for shift(score) info and ctheta for the
   middle.  calls tails for the tails. */

#include "mhead.h"

double middle(info,shift,sufficient,bound,accuracy,nsample,sample,theta,what)
double theta[],info[][NC],shift[],sufficient[][2],bound[],sample[];
int accuracy,nsample[],what;

{
   double qolint[7][NC+1],fun48(),aa[7],numints(),expin2(),d1,e1,cth;
   double fun2(),tails(),numint(),d2,e2;
   int i1,i2,i3,i4,i5,version,version2,k0;
   void intnum2(),intnum3(),intnum4();
/* local: i1,i2,i3,i4,i5 - counter
          polint[]       - integrals of the form x**i.exp(polynomial)         */

/* The tails can be computed analytically                                     */
   cth = tails(info,shift,coef,bound,knots,theta,nknots,what);

/* what == 0: we are only interested in ctheta                                */
   if(what==0){
      for(i1=1; i1<nknots; i1++){               
         for(i2=1; i2<5; i2++){
            dfunpar[i2]=0.;
            for(i3=i1-3; i3<i1+1; i3++){
               if(i3>=0 && i3 < nknots-1)
                  dfunpar[i2]=dfunpar[i2]+coef[i3][i2-1][i1]*theta[i3];
            }
         }
         cth = cth + numint(knots[i1-1],knots[i1],fun2,accuracy);
      }
      return cth;
   }

/* The integrals are computed numerically per interval between knots          */
   for(i1=1;i1<nknots;i1++){               

/* polint gets the result of a numerical integration of the density
   times x^i2                                                                 */
      numints(aa,knots[i1-1],knots[i1],fun48,accuracy,7);
      for(i2=0;i2<7;i2++)qolint[i2][i1]=aa[i2];

/* For the scorefunction we need integrals of a basisfunction times
   the density, this is a combination of qolint[]. Basisfunctions
   do not exists in all intervals.                                            */
      for(i2=i1-3;i2<=i1;i2++){
         if(i2>=0 && i2 < nknots-1 && (i2!=0 || i1!=3)){
            shift[i2] = shift[i2] + qolint[0][i1] * coef[i2][0][i1]
                                  + qolint[1][i1] * coef[i2][1][i1]
                                  + qolint[2][i1] * coef[i2][2][i1]
                                  + qolint[3][i1] * coef[i2][3][i1];

/* For the information matrix we need integrals of a basisfunction
   times a basisfunction times the density, this is a combination of
   qolints. Many combinations do not exist.                                   */
            if(what==2){
               for(i3=i1-3;i3<=i2;i3++)
                  if(i3>=0 && i3 < nknots-1 && (i3!=0 || i1!=3))
                     for(i5=0;i5<4;i5++)
                        for(i4=0;i4<4;i4++)
                           info[i2][i3] = info[i2][i3] + qolint[i5+i4][i1]
                                     * coef[i2][i5][i1] * coef[i3][i4][i1];     
            }
         }
      }
   }
/* The following lines make the informationmatrix and score function
   from what is stored into score and info up to now.                         */
   if(what==2){
      for(i1=0;i1<nknots-1;i1++){
         for(i2=0;i2<=i1;i2++){
            info[i1][i2] = info[i1][i2] - shift[i1] * shift[i2];
            info[i2][i1] = info[i1][i2];
         }
      }
   }
/* if there is no censoring, we are done                                      */
   if(nsample[0]!=nsample[1]){

      version=3;
      if(bound[3] > 0.5) version=4;
      version2=1;
      if(bound[1] > 0.5) version2=2;
      d1 = theta[nknots-2] * coef[nknots-2][1][nknots];
      d2 = theta[0] * coef[0][1][0];
      e2 = theta[0] * coef[0][0][0] -log(ctheta);
      if(nknots>3)
         e1 = theta[nknots-2] * coef[nknots-2][0][nknots]
            + theta[nknots-3] * coef[nknots-3][0][nknots]-log(ctheta);
      else e1 = theta[nknots-2] * coef[nknots-2][0][nknots]-log(ctheta);
      for(i2=0;i2<7;i2++)aa[i2]=0.;
      for(i2=0;i2<7;i2++){
         aa[i2]=1.;
         if(i2!=0)aa[i2-1]=0.;
         qolint[i2][nknots]=expin2(version,knots[nknots-1],bound[4],aa,d1,e1);
      }
      for(i2=0;i2<7;i2++)aa[i2]=0.;
      for(i2=0;i2<7;i2++){
         aa[i2]=1.;
         if(i2!=0)aa[i2-1]=0.;
         qolint[i2][0]=expin2(version2,knots[0],bound[2],aa,d2,e2);
      }
   
      if(accuracy!=0){
         if(nsample[3]>0){
            k0=1;
            for(i2=0;i2<nsample[3];i2++){
            if(i2==nsample[3]-1||sample[i2+nsample[1]+2*nsample[2]]!=
            sample[i2+nsample[1]+2*nsample[2]+1]){
            intnum3(sample[i2+nsample[1]+2*nsample[2]],
                   qolint,d1,e1,version,bound[4],shift,info,k0,nsample[0],what);
            k0=0;
            }
            k0++;
            }
         }
         if(nsample[4]>0){
            k0=1;
            for(i2=0;i2<nsample[4];i2++){
            if(i2==nsample[4]-1||sample[i2+nsample[1]+2*nsample[2]+nsample[3]]!=
            sample[i2+nsample[1]+2*nsample[2]+nsample[3]+1]){
            intnum4(sample[i2+nsample[1]+2*nsample[2]+nsample[3]],
                  qolint,d2,e2,version2,bound[2],shift,info,k0,nsample[0],what);
            k0=0;
            }
            k0++;
            }
         }
         if(nsample[2]>0){
            k0=1;
            for(i2=0;i2<nsample[2];i2++){
               i1=nsample[1]+2*i2;
               if(i2==nsample[2]-1 || sample[i1]!=sample[i1+2] ||
                  sample[i1+1]!=sample[i1+3]){
                  intnum2(sample[i1],sample[i1+1],qolint,shift,info,
                                                            k0,nsample[0],what);
                  k0=0;
               }
               k0++;
            }
         }
      }
      else{
         if(nsample[3]>0){
            i1=0;
            for(i2=0;i2<NC;i2++){
               if(ng3[i2]>0){
                  i1=i1+ng3[i2];
                  intnum3(xg[i2],qolint,d1,e1,version,bound[4],shift,info,
                          ng3[i2],nsample[0],what);
                  if(i1==nsample[3])i2=NC+3;
               }
            }
         }
         if(nsample[4]>0){
            i1=0;
            for(i2=0;i2<NC;i2++){
               if(ng4[i2]>0){
                  i1=i1+ng4[i2];
                  intnum4(xg[i2],qolint,d2,e2,version2,bound[2],shift,info,
                          ng4[i2],nsample[0],what);
                  if(i1==nsample[4])i2=NC+3;
               }
            }
         }
      }
   }
   for(i1=0;i1<nknots-1;i1++)
      shift[i1] = sufficient[i1][0] - shift[i1];
   if(accuracy==0 && nsample[2]>0)
      for(i1=0;i1<nknots-1;i1++)
         shift[i1] = sufficient[i1][1] + shift[i1];
   return ctheta;
}

/******************************************************************************/
void intnum2(x1,x2,qolint,shift,info,n1,n2,what)
double x1,x2,qolint[][NC+1],shift[],info[][NC];
int n2,what,n1;
{
   int i1,i2,i3,i4,i5,jl,jr,im;
   double fun48(),numints(),z0,z1[NC],z2[NC][NC],y1[7],y2[7];

/* jl - in which interval is the left endpoint
   jr - in which interval is the right endpoint                               */

   im = 3*what+1;

/* we are dealing here with one interval censored point - the lower and the
   upper bound can either be in the same interval (i2==i3) or not. In the first
   case we have to compute one set of extra integrals, in the second case two.*/
   for(jl=0; knots[jl]<x1 && jl < nknots; jl++);
   for(jr=0; knots[jr]<x2 && jr < nknots; jr++);
   for(i4=0;i4<7;i4++)y1[i4]=0;
   for(i4=0;i4<7;i4++)y2[i4]=0;
   if(jl==jr){
      numints(y1,x1,x2,fun48,0,im);
   }
   else{
      numints(y1,x1,knots[jl],fun48,0,im);
      numints(y2,knots[jr-1],x2,fun48,0,im);
   }

/* z1 is going to be the first order contribution, z2 the second order, z0
   the normalizing constant. Here we initialize them                          */
   z0=0.;
   for(i1=0;i1<nknots;i1++){
      z1[i1]=0.;
      if(what==2)for(i2=0;i2<nknots;i2++)z2[i1][i2]=0.;
   }

/* for each interval strictly between jl and jr we now compute them           */
   for(i1=1; i1<nknots; i1++){
     if(i1>jl && i1<jr) z0 = z0 + qolint[0][i1];
     for(i3=i1-3; i3<=i1 && i3<nknots-1; i3++) if(i3>=0){
	if(i1>jl && i1<jr) 
           for(i2=0; i2<4; i2++)
	      z1[i3] = z1[i3] + coef[i3][i2][i1] * qolint[i2][i1];
        if(what==2){
           for(i4=i1-3; i4<=i1 && i4<nknots-1;i4++) 
              if(i4>=0 && i1>jl && i1<jr)
                 for(i2=0; i2<4; i2++)
                    for(i5=0;i5<4;i5++)
	               z2[i3][i4] = z2[i3][i4] + coef[i3][i2][i1]
	                             * qolint[i2+i5][i1] * coef[i4][i5][i1];
        }
     }
   }
   z0 = z0 + y1[0] + y2[0];

/* we now compute them for the interval in which the left endpoint is         */
   for(i3=jl-3; i3<=jl && i3<nknots-1; i3++) if(i3>=0){
      for(i2=0; i2<4; i2++)
         z1[i3] = z1[i3] + coef[i3][i2][jl] * y1[i2];
      if(what  == 2){
         for(i4=jl-3; i4<=jl && i4<nknots-1; i4++) 
            if(i4>=0)
               for(i2=0;i2<4;i2++)
                  for(i5=0;i5<4;i5++) 
                     z2[i3][i4] = z2[i3][i4] + coef[i3][i2][jl]
                                   * y1[i2+i5] * coef[i4][i5][jl];
      } 
   } 
/* we now compute them for the interval in which the right endpoint is - since
   y2 = 0 if this interval is the same as the left one, we do not count double*/
   for(i3=jr-3; i3<=jr && i3<nknots-1; i3++) if(i3>=0){
      for(i2=0; i2<4; i2++)
         z1[i3] = z1[i3] + coef[i3][i2][jr] * y2[i2];
      if(what  == 2){
         for(i4=jr-3; i4<=jr && i4<nknots-1; i4++)
            if(i4>=0)
               for(i2=0;i2<4;i2++)
                  for(i5=0;i5<4;i5++) 
                     z2[i3][i4] = z2[i3][i4] + coef[i3][i2][jr]
                                   * y2[i2+i5] * coef[i4][i5][jr];
      } 
   } 

/* now we update shift and info                                               */
   if(z0>0.){
      for(i3=0; i3<nknots; i3++){
	 shift[i3] = shift[i3] - n1 * z1[i3] / (z0 * n2);
         if(what == 2){
	    for(i4=0; i4<nknots; i4++)
               info[i3][i4] = info[i3][i4]+
	          n1 * (z1[i3]*z1[i4] / z0 - z2[i3][i4]) / (n2*z0);
         }
      }
   }
}
/******************************************************************************/
void intnum3(x,qolint,d1,e1,vs,bd,shift,info,n1,n2,what)
double x,qolint[][NC+1],d1,e1,bd,shift[],info[][NC];
int vs,n1,n2,what;
{
   int i1,i2,i3,i4,i5,jin,im;
   double fun48(),numints(),aa[7],z0,yy[7],expin(),z1[NC],z2[NC][NC];
   double expin2();

   im = 3*what+1;

/* in which interval is the point                                             */
   for(jin=0; jin<7; jin++) aa[jin]=0.;
   for(jin=0; jin<7; jin++) yy[jin]=0.;
   for(jin=0; knots[jin]<x && jin < nknots; jin++);
   if(jin<nknots)
      numints(yy,x,knots[jin],fun48,0,im);
   else{
      for(i1=0; i1<7; i1++){
         aa[i1]=1.;
         if(i1!=0) aa[i1-1]=0.;
         yy[i1]=expin2(vs,x,bd,aa,d1,e1);
      }
   }

/* initialize                                                                 */
   z0=0.;
   for(i1=0; i1<nknots; i1++){
      z1[i1]=0.;
      if(what==2) for(i2=0; i2<nknots; i2++) z2[i1][i2]=0.;
   }

/* combine for the intervals which are strictly right of the observation      */
   for(i1=1; i1<=nknots; i1++){
      if(i1>jin) z0=z0+qolint[0][i1];
      for(i3=i1-3; i3<=i1 && i3<nknots-1; i3++) if(i3>=0){
         if(i1>jin)
            for(i2=0; i2<4; i2++)
               z1[i3]=z1[i3]+coef[i3][i2][i1]*qolint[i2][i1];
         if(what==2){
            for(i4=i1-3; i4<=i1 && i4<nknots-1; i4++)
               if(i4>=0 && i1>jin)
                  for(i2=0; i2<4; i2++)
                     for(i5=0; i5<4; i5++)
                        z2[i3][i4] = z2[i3][i4] + coef[i3][i2][i1]
                                   * qolint[i2+i5][i1] * coef[i4][i5][i1];
         }
      }
   }

/* add the one in the interval in which the point is                          */
   z0=z0+yy[0];
   for(i3=jin-3; i3<=jin && i3<nknots-1; i3++) if(i3>=0){
      for(i2=0; i2<4; i2++) z1[i3] = z1[i3] + coef[i3][i2][jin] * yy[i2];
      if(what == 2){
         for(i4=jin-3; i4<=jin && i4<nknots-1; i4++)
            if(i4>=0)
               for(i2=0; i2<4; i2++)
                  for(i5=0; i5<4; i5++)
                     z2[i3][i4] = z2[i3][i4] + coef[i3][i2][jin] * yy[i2+i5]
                                  * coef[i4][i5][jin];
      }
   } 

/* update shift and info                                                      */
   if(z0>0.){
      for(i3=0; i3<nknots-1; i3++){
         shift[i3] = shift[i3] - n1 * z1[i3] / (z0 * n2);
         if(what == 2){
            for(i4=0; i4<nknots-1; i4++) info[i3][i4] = info[i3][i4] 
                         + n1 * (z1[i3] * z1[i4] / z0 - z2[i3][i4]) / (n2 * z0);
         }
      }
   }
}
/******************************************************************************/
void intnum4(x,qolint,d1,e1,vs,bd,shift,info,n1,n2,what)
double x,qolint[][NC+1],d1,e1,bd,shift[],info[][NC];
int vs,n1,n2,what;
{
   int i1,i2,i3,i4,i5,jin,im;
   double fun48(),numints(),aa[7],z0,yy[7],expin(),z1[NC],z2[NC][NC];
   double expin2();

   im = 3*what+1;

/* in which interval is the point                                             */
   for(jin=0; jin<7; jin++) aa[jin]=0.;
   for(jin=0; jin<7; jin++) yy[jin]=0.;
   for(jin=0; knots[jin]<x && jin < nknots; jin++);
   if(jin>0)
      numints(yy,knots[jin-1],x,fun48,0,im);
   else{
      for(i1=0; i1<7; i1++){
         aa[i1]=1.;
         if(i1!=0) aa[i1-1]=0.;
         yy[i1]=expin2(vs,x,bd,aa,d1,e1);
      }
   }

/* initialize                                                                 */
   z0=0.;
   for(i1=0; i1<nknots; i1++){
      z1[i1]=0.;
      if(what==2) for(i2=0; i2<nknots; i2++) z2[i1][i2]=0.;
   }

/* combine for the intervals which are strictly right of the observation      */
   for(i1=0; i1<nknots; i1++){
      if(i1<jin) z0=z0+qolint[0][i1];
      for(i3=i1-3; i3<=i1 && i3<nknots-1; i3++) if(i3>=0){
         if(i1<jin)
            for(i2=0; i2<4; i2++)
               z1[i3]=z1[i3]+coef[i3][i2][i1]*qolint[i2][i1];
         if(what==2){
            for(i4=i1-3; i4<=i1 && i4<nknots-1; i4++)
               if(i4>=0 && i1<jin)
                  for(i2=0; i2<4; i2++)
                     for(i5=0; i5<4; i5++)
                        z2[i3][i4] = z2[i3][i4] + coef[i3][i2][i1]
                                   * qolint[i2+i5][i1] * coef[i4][i5][i1];
         }
      }
   }

/* add the one in the interval in which the point is                          */
   z0=z0+yy[0];
   for(i3=jin-3; i3<=jin && i3<nknots-1; i3++) if(i3>=0){
      for(i2=0; i2<4; i2++) z1[i3] = z1[i3] + coef[i3][i2][jin] * yy[i2];
      if(what == 2){
         for(i4=jin-3; i4<=jin && i4<nknots-1; i4++)
            if(i4>=0)
               for(i2=0; i2<4; i2++)
                  for(i5=0; i5<4; i5++)
                     z2[i3][i4] = z2[i3][i4] + coef[i3][i2][jin] * yy[i2+i5]
                                  * coef[i4][i5][jin];
      }
   } 

/* update shift and info                                                      */
   if(z0>0.){
      for(i3=0; i3<nknots-1; i3++){
         shift[i3] = shift[i3] - n1 * z1[i3] / (z0 * n2);
         if(what == 2){
            for(i4=0; i4<nknots-1; i4++) info[i3][i4] = info[i3][i4] 
                         + n1 * (z1[i3] * z1[i4] / z0 - z2[i3][i4]) / (n2 * z0);
         }
      }
   }
}
/******************************************************************************/
double tails(info,shift,coef,bound,knots,theta,nknots,what)

/* tails1 computes the integrals in the tails.                                */

double info[][NC],shift[],coef[][4][NC],bound[],theta[],knots[];
int nknots,what;

{
   double a[6],expin(),cth;
   int i,j,version;
/* local:
   a,version:    for expin
   i,j      :    counters                                                     */
      
/* Initialize cth, score and info                                             */
   cth = 0.;
   if(what>=1){
      for(i=0;i<nknots;i++){
         shift[i] = 0.;
         if(what==2) for(j=0;j<nknots;j++) info[i][j]=0.;
      }
   }

/* Now we are going to compute the integrals, first the ones  
   for the left tail, then the right tail, a large number of   
   integrals are zero, and they do not have to be computed  

   First the left-tail, only the first basisfunction is not
   equal to 0 here. It only has a linear and a constant term
   in the tails.                                                              */

   version=1;
   if(bound[1] > 0.5) version=2;
      
/* The numbers a1,b1,c1,d1,e1 are constants in the integrals                  */
   a[1] = 0.;
   a[4] = theta[0] * coef[0][1][0];
   a[5] = theta[0] * coef[0][0][0];
   if(what==0){
      a[2] = 0.;
      a[3] = 1.;
      cth = expin(version,knots[0],bound[2],a);
   }
   else{
      a[2] = coef[0][1][0];
      a[3] = coef[0][0][0];
      shift[0] = expin(version,knots[0],bound[2],a) /ctheta;
      if(what==2){
         a[1]=coef[0][1][0]*coef[0][1][0];
         a[2]=coef[0][1][0]*coef[0][0][0]*2.;
         a[3]=coef[0][0][0]*coef[0][0][0];
         info[0][0] =expin(version,knots[0],bound[2],a)/ctheta;
      }
   } 
      
/* Now the right tail. Only the last 2 basisfunctions are not
   equal to 0 here. The last one has a linear and a constant
   term, the one but last one only a constant term.                           */
   version=3;
   if(bound[3] > 0.5) version=4;
   a[1] = 0.;
   a[4] = theta[nknots-2] * coef[nknots-2][1][nknots];
   if(nknots>3)
      a[5] = theta[nknots-2] * coef[nknots-2][0][nknots]
           + theta[nknots-3] * coef[nknots-3][0][nknots];
   else a[5] = theta[nknots-2] * coef[nknots-2][0][nknots];
   if(what == 0){
      a[2] = 0.;
      a[3] = 1.;
      cth = cth + expin(version,knots[nknots-1],bound[4],a);
   }
   else {
      if(nknots>3){
         a[2] = 0.;
         a[3] = coef[nknots-3][0][nknots];
         shift[nknots-3] = expin(version,knots[nknots-1],bound[4],a)/ctheta;
      }
      a[2] = coef[nknots-2][1][nknots];
      a[3] = coef[nknots-2][0][nknots];
      shift[nknots-2] = expin(version,knots[nknots-1],bound[4],a)/ctheta;
   }
   if(what == 2 && nknots>3){
      a[2] = 0.;
      a[3] = coef[nknots-3][0][nknots] * coef[nknots-3][0][nknots];
      info[nknots-3][nknots-3]=expin(version,knots[nknots-1],bound[4],a)/ctheta;
      a[2] = coef[nknots-3][0][nknots] * coef[nknots-2][1][nknots];
      a[3] = coef[nknots-3][0][nknots] * coef[nknots-2][0][nknots];
      info[nknots-2][nknots-3]=expin(version,knots[nknots-1],bound[4],a)/ctheta;
      info[nknots-3][nknots-2] = info[nknots-2][nknots-3];
   }
   if(what == 2){
      a[1] = coef[nknots-2][1][nknots] * coef[nknots-2][1][nknots];
      a[2] = coef[nknots-2][0][nknots] * coef[nknots-2][1][nknots] *2.;
      a[3] = coef[nknots-2][0][nknots] * coef[nknots-2][0][nknots];
      info[nknots-2][nknots-2]=expin(version,knots[nknots-1],bound[4],a)/ctheta;
   }
   return cth;
}
