/*******************************************************************************
* Copyright (c) 1993, 1995 by Charles Kooperberg.                              *
* All rights reserved. This function is part of the Logspline Density          *
* Estimation Program. It was written by Charles Kooperberg at the University   *
* of Washington and the University of California at Berkeley between 1988 and  *
* 1991. It is described in the paper: Kooperberg, Charles and Stone, Charles J.*
* `Logspline density estimation for censored data', Journal of Computational   *
* and Graphical Statistics, 1 (1992) 301-328.                                  *
* You are free to use this program, for non-commercial purposes only,          *
* under two conditions:                                                        *
* (1) This note is not to be removed                                           *
* (2) Publications using logspline computations should reference the           *
*  publication  mentioned above.                                               *
* For questions, please email clk@stat.washington.edu                          *
*                       Charles Kooperberg, November 30, 1995                  *
*******************************************************************************/
/* this is the main iteration loop                                            */
/* setbounds follows at the end                                               */

#include "mhead.h"

double iter(info,sufficient,bound,SorC,nsample,sample,accuracy,itrouble)

double info[][NC],sufficient[][2],bound[],sample[];
int SorC[],accuracy,nsample[],*itrouble;

{
   int numbertester();
   int counter=0,infol,i1,i2,i3,kpvt[NC],jaja1,i4,iii[4],ithere,i7,nrc=0,nrc2=0;
   double error,oldlikelihood,cbound[7],shift[NC],erroradjust(),dd[2];
   double middle(),candidate[NC],newlikelihood,work[NC][NC],error2();
   double likeli(),one,rvr[100],linsearch(),errorx;
   void setbounds();

/* local
   i1 i2 i3    - counters
   oldlikelihood - loglikelihood previous iteration
   newlikelihood - loglikelihood present iteration
   infol,kpvt  - for linpack
   work        - for linpack
   shift       - used for the shift and score
   error       - stop criterion
   cbound      - see below
   candidate   - candidate for new theta
   counter     - number of iterations since last reset of boundaries          */

   one=0.99999;
/* sets integration bounds                                                    */
   setbounds(bound,cbound,nsample);

/* accuracy = 0 means that we approximate censoring                           */
   if(accuracy == 0 && nsample[0]==nsample[1])accuracy=1;

/* start of the iteration                                                     */
   if(SorC[0] == 0) printf("%d \n",nknots);
   for(i1=1; i1<500; i1++){

/* if we go extremely far out in the tails, and theta[first] or theta[last]
   still doesn't have the right sign, we fix the bounds at the max value      */
      ithere = 0.;
      if(bound[3]<0.5||bound[3]>1.5)ithere=5;
      else{
         if(bound[4]>=25*knots[nknots-1]-24*knots[0]){
            if(SorC[0]==0&&SorC[24]==0)printf("a very long right tail");
            SorC[24]=nknots;
            *itrouble=17;
            for(i2=1;i2<5;i2++)bound[i2]=cbound[i2];
            return 0.;
         }
      }
      if(bound[1]<0.5||bound[1]>1.5)ithere=ithere+5;
      else{
         if(bound[2]<=-24*knots[nknots-1]+25*knots[0]){
            if(SorC[0]==0&&SorC[23]==0)printf("a very long left tail");
            SorC[23]=nknots;
            *itrouble=17;
            for(i2=1;i2<5;i2++)bound[i2]=cbound[i2];
            return 0.;
         }
      }

/* after 250 iterations we do the censoring exact                             */
      if(accuracy == 0 && i1==250){
         accuracy=1;
         counter = 0;
      }
      jaja1=0;
      counter++;

/* if something changed in the bounds since last time counter = 0, and we have 
   to recompute ctheta and the loglikelihood                                  */
      i7=0;
      do{
      if(counter==1){
         do{
            ctheta = middle(info,shift,sufficient,bound,accuracy,
                                                        nsample,sample,theta,0);
/* check against overflow and NAs                                             */
            i2=-1;
            if(ctheta>0.0)i2=i2+1;
            if(ctheta<2.0)i2=i2+1;
            if(numbertester(ctheta)==1)i2=i2-4;
            if(i2<0){
               for(i2=0;i2<nknots-1;i2++)theta[i2]=theta[i2]/10.;
               if(i7>7 && *itrouble == 0){
                  *itrouble = 2;
                  for(i2=1;i2<5;i2++)bound[i2]=cbound[i2];
                  return 0.;
               }
               i2=-6;
               i7++;
               if(i7==200){
                  *itrouble=5;  
                  return 1.;
               }
            }
         }while(i2<0);
         newlikelihood=likeli(theta,nsample,sample,bound,accuracy);
      }
      i2=1000*nknots+i1;
      
/* compute score (stored in shift), info and ctheta                           */
      ctheta = middle(info,shift,sufficient,bound,accuracy,
                                                        nsample,sample,theta,2);
           
/* store the loglikelihood of the previous iteration                          */
      oldlikelihood = newlikelihood;

/* copy info                                                                  */
      for(i2=0;i2<nknots-1;i2++)
         for(i3=0;i3<nknots-1;i3++) work[i2][i3] = info[i2][i3];
/* solve the system                                                           */
      i2=NC;
      i3=nknots-1;
      i4 = 100;
      xdsifa_(work,&i2,&i3,kpvt,&infol);
/* check for underflows overflows and NA                                      */
      if(infol!=0) {for(i2=0;i2<nknots-1;i2++) theta[i2] = theta[i2]/100.;
         if(i7>7 && *itrouble == 0){
            *itrouble = 2;
            for(i3=1;i3<5;i3++)bound[i3]=cbound[i3];
            return 0.;
         }
         i7++;
         if(i7==200){
            *itrouble = 5;
            return 1.;
         }
      }
      }while(infol!=0);
      iii[1]=0;
      iii[2]=0;
      if(nsample[2]+nsample[3]+nsample[4] >0)
         xdsidi_(work,&i2,&i3,kpvt,dd,iii,rvr,&i4); 
      for(i4=0;i4<nknots-1;i4++)rvr[i4]=shift[i4];
      iii[3]=iii[1]+iii[2];
      if(iii[2]+iii[1]==0)nrc=0;
      if(iii[2]+iii[1]==0)nrc2=0;
      if(iii[1]+iii[2]>0&&nrc+nrc2>=5&&nrc>=2){
         nrc2++;
         if(nrc2==15 && *itrouble == 0){
              *itrouble = 2;
              for(i3=1;i3<5;i3++)bound[i3]=cbound[i3];
              return 0.;
          }
         if(nrc2==48 && *itrouble != 0){
              *itrouble = 5;
              for(i3=1;i3<5;i3++)bound[i3]=cbound[i3];
              return 1.;
          }
          iii[2]=0;
          iii[1]=0;
          nrc=0;
      }
      if(iii[2]+iii[1]==0){
         xdsisl_(work,&i2,&i3,kpvt,shift);  

/* compute the stop criterion and adjust stepsize, if too large               */
         i7=0;
         do{
            errorx=error2(shift,rvr);
            errorx=errorx*100.;
            error = erroradjust(shift);
            if(iii[3]==0)error=errorx;
            i2=-1;
            if(error>0.0)i2=i2+2;
            if(error<2.0)i2=i2+2;
            if(numbertester(error)==1)i2=i2-4;
            if(i2<0){
               for(i2=0;i2<nknots-1;i2++)shift[i2]=shift[i2]/10.;
               if(i7>7 && *itrouble == 0){
                  *itrouble = 2;
                  for(i3=1;i3<5;i3++)bound[i3]=cbound[i3];
                  return 0.;
               }
               i7++;
               if(i7==200){
                  *itrouble = 5;
                  return 1.;
               }
               i2=-7;
            }
         }while(i2<0);

/* counter is the number of iterations since the last adjustment of integration
   boundaries. Thus if we find an error, we put it to 0 and start all over.
   Temporarily we deduct 10000 from it if we have to half the step size first */
         i7=0;
         do{
            if(counter<-1000)counter=counter+10000;

/* We should check whether theta's that should be negative (the ones for tail
   basis functions if we are integrating to plus/minus infinity) stay negative.
   If counter become 0 we essentially leave the loop here.....................
   We then go from case A to case B...........................................*/
            if(theta[0] + shift[0] >= 0. && bound[1] < 0.5){
               cbound[5] = -cbound[5];
               bound[2] = cbound[5] * (knots[0] - knots[1]) + knots[0];
               bound[1] = 1;
               counter = 0;
            }
            if(theta[nknots-2] + shift[nknots-2] >= 0. && bound[3] < 0.5){
               cbound[6] = -cbound[6];
               bound[4] = cbound[6] * (knots[nknots-1] - knots[nknots-2]) +
                              knots[nknots-1];
               bound[3] = 1.;
               counter = 0;
            }

/* Compute the C(theta) for the candidate theta-hat. We can
   then compute the new log-likelihood.                                       */
            if(counter >0){
               for(i2=0;i2<nknots-1;i2++) candidate[i2] = shift[i2] + theta[i2];
               ctheta = middle(info,shift,sufficient,bound,accuracy,
                                                    nsample,sample,candidate,0);
               i2=0;
               if(ctheta>0.0)i2=i2+1;
               if(ctheta<2.0)i2=i2+1;
               if(numbertester(ctheta)==1)i2=i2-4;
               if(i2>0){
               newlikelihood=likeli(candidate,nsample,sample,bound,accuracy);

/* If the loglikelihood really decreases, we step size and go back a bit      */
               if(newlikelihood < oldlikelihood &&
                  ((newlikelihood/oldlikelihood < one) ||
                  (oldlikelihood/newlikelihood < one)) && error > 0.00001){
		  jaja1++;
                  if(jaja1<12 || i1<3){
                     for(i2=0;i2<nknots-1;i2++) shift[i2] = shift[i2]/2.;
                     counter = counter - 10000;
                  }
               }
               }
               else{
                  if(i7 > 7 && *itrouble == 0){
                     *itrouble = 2;
                     for(i3=1;i3<5;i3++)bound[i3]=cbound[i3];
                     return 0.;
                  }
                  i7++;
                  if(i7==200){
                     *itrouble = 5;
                     return 1.;
                  }
                  for(i2=0;i2<nknots-1;i2++) shift[i2] = shift[i2]/10.;
                  counter = counter - 10000;
               }
            }
         }while(counter <0);
/* Only if the counter is negative we have to go back.                        */
      }
      if(jaja1==25)for(i3=1;i3<5;i3++)bound[i3]=cbound[i3];
      if(jaja1==25)return newlikelihood;
      if(iii[2]+iii[1] >0){
         nrc++;
         for(i2=0;i2<nknots-1;i2++) shift[i2] = rvr[i2];
         for(i2=0;i2<nknots-1;i2++) rvr[i2] = theta[i2];
         erroradjust(shift);
         if(cbound[5]<0. && bound[1] < 0.5 &&
            (cbound[5]>-6.||theta[0]+shift[0]>0)){
            cbound[5] = -cbound[5];
            bound[2] = cbound[5] * (knots[0] - knots[1]) + knots[0];
            bound[1] = 1;
            counter = 0;
         }
         if(cbound[6]<0. && bound[3] < 0.5 && 
            (cbound[6]>-6.||theta[nknots-2]+shift[nknots-2]>0)){
            cbound[6] = -cbound[6];
            bound[4] = cbound[6] * (knots[nknots-1] - knots[nknots-2]) +
                           knots[nknots-1];
            bound[3] = 1.;
            counter = 0;
         }
         if(counter ==0){
            ctheta = middle(
                   info,shift,sufficient,bound,accuracy,nsample,sample,theta,0);
            middle(info,shift,sufficient,bound,accuracy,nsample,sample,theta,1);
            oldlikelihood=likeli(theta,nsample,sample,bound,accuracy);
            for(i2=0;i2<nknots-1;i2++) shift[i2] = rvr[i2];
            for(i2=0;i2<nknots-1;i2++) rvr[i2] = theta[i2];
            erroradjust(shift);
         }
         oldlikelihood = linsearch(
                             shift,oldlikelihood,bound,nsample,sample,accuracy);
         ctheta = middle(
                   info,shift,sufficient,bound,accuracy,nsample,sample,theta,0);
         middle(info,shift,sufficient,bound,accuracy,nsample,sample,theta,1);
         erroradjust(shift);
         oldlikelihood = linsearch(
                             shift,oldlikelihood,bound,nsample,sample,accuracy);
         for(i2=0;i2<nknots-1;i2++) shift[i2]= theta[i2] - rvr[i2];
         oldlikelihood = linsearch(
                             shift,oldlikelihood,bound,nsample,sample,accuracy);
         ctheta = middle(
                   info,shift,sufficient,bound,accuracy,nsample,sample,theta,0);
         for(i2=0;i2<nknots-1;i2++) shift[i2]= 0.;
         for(i2=0;i2<nknots-1;i2++)rvr[i2]=theta[i2]-rvr[i2];
         error = 100.*erroradjust(rvr);
         newlikelihood=oldlikelihood;
      }

/* O.K. the new theta is good, let's store it.                                */
      if(counter > 0){
         for(i2=0;i2<nknots-1;i2++) theta[i2] = theta[i2] + shift[i2];

/* Is there convergence? If this is not the final iteration, we're not picky. */
         if(ithere<10) error=error/100.;
         if(error < 0.000001){

/* If we were integrating to +/-infinity or if we exceeded the bounds, done.  */
            if(ithere >9){
                if(accuracy == 5){
                   for(i3=1;i3<5;i3++)bound[i3]=cbound[i3];
                   rvr[99] = middle(
                          info,shift,sufficient,bound,5,nsample,sample,theta,0);
                   if(iii[1]+iii[2]!=0)
                   rvr[99] = middle(
                          info,shift,sufficient,bound,5,nsample,sample,theta,0);
                   i2=0;
                   if(rvr[99]>0.0)i2=i2+1;
                   if(rvr[99]<2.0)i2=i2+1;
                   if(numbertester(rvr[99])==1)i2=i2-4;
                   if(i2>0){
                      ctheta=rvr[99];
                      newlikelihood=likeli(theta,nsample,sample,bound,5);
                   }
                   return newlikelihood;
                } 
                else {
                   accuracy = 5;
                   counter = 0;
                }
            }
            else {
          
/* If we were not integrating to +/- infinity, we either double the
   integration tails (if a tail theta was possitive) or we go for it:
   +/- infinity (That means cbound[5] and cbound[6] become negative.)         */
               counter=0;
/* Left tail                                                                  */
               if(cbound[5] < 0)                                /* case A     */
                  cbound[5] = 2. * cbound[5];
               else{        
                  if(cbound[1] < 0.5 && theta[0] >=0.){         /* case B     */
                     cbound[5] = 2. * cbound[5];
                     bound[2] = cbound[5] * (knots[0] - knots[1]) + knots[0];
                  }
                  if(cbound[1] < 0.5 && theta[0] <0. ){         /* case B to A*/
                     cbound[5] = -2. * cbound[5];
                     bound[1] = 0;
                  }
                  if(cbound[1] > 0.5 && bound[1] > 0.5 && bound[1] < 1.5){ /*E*/
                     cbound[5] = 2. * cbound[5];
                     bound[2] = cbound[5] * (knots[0] - knots[1]) + knots[0];
                     if(bound[2] < cbound[2]){                  /* goto C/D   */
                        bound[2] = cbound[2];
                        bound[1] = 2.;   
                     }
                  }
               }
/* Right tail                                                                 */
               if(cbound[6] < 0)                                /* case A     */
                  cbound[6] = 2. * cbound[6];
               else{        
                  if(cbound[3] < 0.5 && theta[nknots-2] >=0.){  /* case B     */
                     cbound[6] = 2. * cbound[6];
                     bound[4] = cbound[6] * (knots[nknots-1] - knots[nknots-2])
                                                             + knots[nknots-1];
                  }
                  if(cbound[3] < 0.5 && theta[nknots-2] <0. ){  /* case B to A*/
                     cbound[6] = -2. * cbound[6];
                     bound[3] = 0.;
                  }
                  if(cbound[3] > 0.5 && bound[3] > 0.5 &&
                                                    bound[3] < 1.5){ /* case E*/
                     cbound[6] = 2. * cbound[6];
                     bound[4] = cbound[6] * (knots[nknots-1] - knots[nknots-2])
                                                             + knots[nknots-1];
                     if(bound[4] > cbound[4]){                  /* goto C/D   */
                        bound[4] = cbound[4];
                        bound[3] = 2;   
                     }
                  }
               }
            }
         }
      }
   }

/* If we ended up here, there was no convergence in 300 iterations.           */
   if(SorC[0] == 0){
      printf("no convergence was achieved with %d knots\n",nknots);
         SorC[0] = -647;
   }
   else
      SorC[0] = -SorC[0];
   return newlikelihood;
}
void setbounds(bound,cbound,nsample)
double bound[],cbound[];
int nsample[];
{
/* set the integration boundaries. O.k., this is quite complicated. There are 2
   arrays that determine how far we are integrating in each tail.
   For the lower tail:

   bound[1] - is there a lower bound to which we should integrate (on entry)
          are we right now integrating to a lower bound (during)
          is there a lower bound to which we should integrate (on exit)
          (0 = -infinity; 1 = lower bound);
   cbound[1]- copies the begin value of bound[1] for during the iterations.
   bound[2] - if bound[1] == 1/2 the lower bound of integration (on entry)
          if bound[1] == 1/2 the lower bound of integration (during)
          if bound[1] == 1/2 the lower bound of integration (on exit)
   cbound[2]- copies the begin value of bound[2] for during the integration.
   cbound[5] - technical limit on the integration. If cbound[5] < 0 we can
          integrate to -infinity (if we want to), if cbound[5] > 0 the
          furthest we want to integrate is to
          cbound[5]*(knots[1]-knots[2])+knots[1];
          if cbound[5] is smaller than 0 it is twice its last possitive value.
   For the upper tail: bound[3], bound[4], cbound[3], cbound[4], cbound[6]    */

/* the lower tail                                                             */
   cbound[2] = bound[2];
   cbound[1] = bound[1];
   cbound[5] = 1.;
   if(nsample[3]*3>nsample[1])cbound[5]=0.5;
   bound[2] = cbound[5] * (knots[0] - knots[1]) + knots[0];

/* there are 5 possible situations:
A  cbound[1]=0, theta[0]<0    need: cbound[5]=-1, bound[1]=0
B               theta[0]>=0   need: cbound[5]=1,  bound[1]=1
C  cbound[1]=1, theta[0]<0    need: cbound[5]=1,  bound[1]=2, bound[2]=cbound[2]
D               theta[0]>=0, bound[2]<cbound[2]:
                                     cbound[5]=1, bound[1]=2, bound[2]=cbound[2]
E                            bound[2]>=cbound[2]:
                                     cbound[5]=1,  bound[1]=1              */
   if(theta[0]<0. && cbound[1] < 0.5 && nsample[0] == nsample[1])
      cbound[5] = -1.;        
   if(cbound[1] < 0.5 && (theta[0] >= 0. || nsample[0]!=nsample[1]))
      bound[1] = 1;
   if(cbound[1] > 0.5) bound[1] = 2;
   if(cbound[1] > 0.5 && theta[1] >= 0. && bound[2] >= cbound[2]) bound[1] = 1;
   if(cbound[1] > 0.5 && (theta[0]<0. || (theta[0] >= 0. && 
                   bound[2] < cbound[2]))) bound[2] = cbound[2];

/* the upper tail almost similar                                              */
   cbound[4] = bound[4];
   cbound[3] = bound[3];
   cbound[6] = 1;
   if(nsample[4]*3>nsample[1])cbound[6]=0.5;
   bound[4] = cbound[6] * (knots[nknots-1] - knots[nknots-2]) + knots[nknots-1];
   if(theta[nknots-2]<0. && cbound[3] < 0.5 && nsample[0] == nsample[1])
      cbound[6] = -cbound[6]; 
   if(cbound[3] < 0.5 && (theta[nknots-2] >=0. || nsample[0]!=nsample[1]))
      bound[3] = 1;
   if(cbound[3] > 0.5) bound[3] = 2;
   if(cbound[3] > 0.5 && theta[nknots-2] >= 0. && cbound[4] >= bound[4])
      bound[3] = 1;
   if(cbound[3] > 0.5 && (theta[nknots-2]<0. || (theta[nknots-2] >= 0. &&
              bound[4] > cbound[4]))) bound[4] = cbound[4];
      
}
