/*******************************************************************************
* Copyright (c) 1993, 1995 by Charles Kooperberg.                              *
* All rights reserved. This function is part of the Logspline Density          *
* Estimation Program. It was written by Charles Kooperberg at the University   *
* of Washington and the University of California at Berkeley between 1988 and  *
* 1991. It is described in the paper: Kooperberg, Charles and Stone, Charles J.*
* `Logspline density estimation for censored data', Journal of Computational   *
* and Graphical Statistics, 1 (1992) 301-328.                                  *
* You are free to use this program, for non-commercial purposes only,          *
* under two conditions:                                                        *
* (1) This note is not to be removed                                           *
* (2) Publications using logspline computations should reference the           *
*  publication  mentioned above.                                               *
* For questions, please email clk@stat.washington.edu                          *
*                       Charles Kooperberg, November 30, 1995                  *
*******************************************************************************/
/* this is the main program                                                   */
/* remove follows at the end                                                  */

#include "mhead.h"
int logcensor(idelete,iknotauto,sample,nsample,bound,
              SorC,ynknots,yknots,ycoef,alpha,wk,wk2,logl)

int *idelete,*iknotauto,nsample[],SorC[],*ynknots;
double bound[],sample[],yknots[],wk[],wk2[],ycoef[],*alpha,logl[];
/* these quantities are defined in the file where they originated and lhead.h */

{
   int removeknot(),knotnumber(),itrouble,accuracy=0,xnknots;
   double iter(),qt[2];
   void fits();
/* functions - see the functions themselves */

double info[NC][NC],loglikelihood,sufficient[NC][2],coef2[NC][NC],xctheta;
/* stuf used in iter
   info - the information matrix
   loglikelihood                                  
   sufficient - sufficient statistics
   coef,coef2 - 2 matrices defining the splines as a function of the knots, see
                lcoef.c for the exact definitions.                            */

double derivatives[NC],crossprods[NC][NC],suffcombine[NC][2],xtheta[NC];
/* derivative crossprods used to compute the starting values - see lstart.c    
   suffcombine used to compute the sufficient statistics - see lsuff.c        */

double aic,aicmin,r1,rknots[NC],xcoef2[NC][NC];
/* local double stuff:
   r1 - utility
   aic - akaike information criterion
   aicmin - minimum aic encountered
   alpha - alpha value in aic               
   rknots - copy of knots                                                     */

int i,j,nkstart,iremove=0,iknots2[NC],iknots[NC],xiknots[NC];
void coeff() ,start1() ,start2() ,suffstat1() ,suffstat2() ,knotplace();
/* local integers
   i,j,k - counter, utility                                                   
   nkstart - number of knots at the beginning of the algorithm
   iknots2 - copy of iknots
   iremove - number of the knot that is removed                               */



/******************************************************************************/

/* compute the number of knots (to start)                                     */
   for(i=0;i<NC;i++) ng3[i]=0;
   for(i=0;i<NC;i++) ng4[i]=0;
   nknots = *ynknots;
   for(i=0;i<NC;i++)knots[i]=yknots[i];

/* 14 is the minimum sample size                                              */
   if(nsample[0] < 14){
      if(SorC[0] == 0)
         printf("sample is too small\n");
      else
         SorC[0] = 2;
      return 0;
   }

/* determine the number of starting knots                                     */
   nknots = knotnumber(*idelete,nsample,nknots,SorC);

/* place the knots                                                            */
   knotplace(iknots,rknots,*iknotauto,bound,sample,nsample,SorC,wk,wk2,qt);
   bound[2]=qt[0]+qt[1]*bound[2];
   bound[4]=qt[0]+qt[1]*bound[4];

/* initialize aicmin to a very large value                                    */
   aicmin = pow(10.,100.);

   nkstart = nknots;
   for(i=0;i<nknots;i++)xg[i]=knots[i];
   for(i=nsample[1]+2*nsample[2];i<nsample[0]+nsample[2]-nsample[4];i++){
      for(j=0;knots[j]<sample[i]&&j<nknots;j++);
      if(j!=0&&j!=nknots)
         if(sample[i]-knots[j-1]<knots[j]-sample[i])j=j-1;
      if(j==nknots)j=j-1;
      ng3[j]++;
   }
   for(i=nsample[1]+2*nsample[2]+nsample[3];i<nsample[0]+nsample[2];i++){
      for(j=0;knots[j]<sample[i]&&j<nknots;j++);
      if(j!=0&&j!=nknots)
         if(sample[i]-knots[j-1]<knots[j]-sample[i])j=j-1;
      if(j==nknots)j=j-1;
      ng4[j]++;
   }
   for(i=0;i<nknots;i++)yknots[i]=knots[i];

/* some possible errors                                                       */
   if(SorC[0] == -2)return 0;
   if(SorC[0] == 647)return 0;
   if(SorC[0] == 23)return 0;

/* Compute stuff later used for computing sufficient statistics.              */
   suffstat1(suffcombine,sample,nsample);

/* Compute stuff later used for computing the starting values.                */
   start1(crossprods,derivatives,sample,nsample);

/* Compute coefficient matrix.                                                */
   itrouble = 0;
   do{   
      coeff(coef2);

/* Compute sufficient statistics.                                             */
      suffstat2(suffcombine,coef2,sufficient);

/* Compute starting values                                                    */
      start2(crossprods,derivatives,coef2,nkstart,iremove);

/* Compute theta-hat (main iteration loop).                                   */
      if(itrouble == 1){
         for(i=0;i<nknots-1;i++)theta[i]=0.;
         loglikelihood = iter(info,sufficient,bound,SorC,nsample,sample,5,
         &itrouble);
      }
      else{
      loglikelihood = iter(info,sufficient,bound,SorC,nsample,sample,accuracy,
         &itrouble);
      if(itrouble == 2){
         for(i=0;i<nknots-1;i++)theta[i]=0.;
         itrouble = 1;
         loglikelihood = iter(info,sufficient,bound,SorC,nsample,sample,5,
         &itrouble);
      }}

/* this means that there were problems (like no convergence)                 */
      SorC[27]=0;
      if(SorC[0] < 0 || itrouble == 5||itrouble==17){
         if(nknots==nkstart){
             SorC[0] = -1;
             SorC[27]=0;
             return 0;
         }
         if(SorC[0]==-647)
            printf("Smallest number of knots tried is %d\n",nknots+1);
         SorC[27]=nknots;
         nknots=0;
      }
      
/* If we have a fixed number of knots we are done now.           
   Copy the solution in the x-arrays                                         */
      if(nknots>0){
      if(*idelete == 0){
         xctheta = ctheta;
         xnknots = nknots;
         for(i=0; i<NC; i++){
            xtheta[i] = theta[i];
            xiknots[i] = iknots[i];
            for(j=0; j<NC; j++)
               xcoef2[i][j] = coef2[i][j];
         }
         logl[0]=(loglikelihood+log(qt[1]))*nsample[0];
      }
   
/* We are deleting knots. Compute aic. Is it an improvement?                  */
      else {
         aic = -2. * loglikelihood * nsample[0] + *alpha * (nknots-1);      
         logl[nknots-1]=(loglikelihood+log(qt[1]))*nsample[0];
         if(aic <= aicmin){

/* Then we mark the solution iknots2 is put to 0, so we can fill it later     */
            aicmin = aic;
            xctheta = ctheta;
            xnknots = nknots;
            for(i=0; i<NC; i++){
               xtheta[i] = theta[i];
               xiknots[i] = iknots[i];
               iknots2[i] = 0;
               for(j=0;j<NC;j++)
                  xcoef2[i][j]=coef2[i][j];
            }

/* that is we fill it here, iknots contains the rank numbers of the remaining
   knots                                                                      */
            for(i=0;i<nknots;i++)
               iknots2[iknots[i]] = 1;
         }
         else{
            r1 = -2. * loglikelihood * nsample[0] + *alpha * 2;
            if(r1 >= aicmin) nknots = 0;
         }

      
/* If there were more than 4 knots, we should remove one.                     */
         if(nknots >= 4){

/* Select the one to remove.                                                  */
            iremove = removeknot(info,coef2);

/* Remove its remainders from all sort of arrays. That is, shift the ends of the
   array 1 closer to 0, crossprods gets shifted in 2 directions               */
            for(i=iremove; i<nknots; i++){
               knots[i] = knots[i+1];
               iknots[i] = iknots[i+1];
               derivatives[i+1] = derivatives[i+2];
               suffcombine[i+2][0] = suffcombine[i+3][0];
               suffcombine[i+2][1] = suffcombine[i+3][1];
               crossprods[0][i+1] = crossprods[0][i+2];
               crossprods[i+1][0] = crossprods[i+2][0];
            }
            for(j=0; j<nknots; j++){
               for(i=iremove; (i<nknots && j>=iremove);i++)
                  crossprods[i+1][j+1] = crossprods[i+2][j+2];
               for(i=iremove;(i<nknots && j<iremove);i++)
                  crossprods[i+1][j+1] = crossprods[i+2][j+1];
               for(i=0;(i<iremove && j>=iremove);i++)
                  crossprods[i+1][j+1] = crossprods[i+1][j+2];
            }
         }

/* this takes care that we stop                                               */
         else
           if(nknots == 3)nknots=2;
      }}
   } while(nknots>=3 && *idelete >0);

/* Now write the solution down.   This prints the density                     */
   nknots = nkstart;
   fits(xcoef2,xtheta,xctheta,ycoef,xiknots,xnknots);
   *ynknots = nkstart;
   bound[2]=(bound[2]-qt[0])/qt[1];
   bound[4]=(bound[4]-qt[0])/qt[1];
   for(i=0;i<nkstart;i++)
       yknots[i]=(yknots[i]-qt[0])/qt[1];
   for(i=0;i<nkstart;i++)
       ycoef[i+2]=ycoef[i+2]*qt[1]*qt[1]*qt[1];
   ycoef[0]=ycoef[0]+ycoef[1]*qt[0]+log(qt[1]);
   ycoef[1]=ycoef[1]*qt[1];
   
   return nkstart;
}
/******************************************************************************/

int removeknot(info,coef2)

double info[][NC],coef2[][NC];
/* all described in originating files and lhead.h                             */

{
   double ratmax=0.,wk1[NC],se[NC],phi[NC],x1[2];
   int i,j,infol,k,x2[3],x3=1,kpvt[NC],irmax=1;
/* local:

	   wk1[]      - workarray for linpack
	   kpvt       - workarray for linpack
	   x1,x2,x3   - for linpack
	   infol      - for linpack
	   i j        - counters
	   phi        - linear combination of thetas
	   se         - standard errors of phi
	   ratmax     - maximum ratio se/phi

	   output:

	   irmax      - index of maximum ratio                                        */

	/* Take linear combinations of theta such that phi is theta(phi) for
	   the truncated power basis. (Which is not a basis.)                         */
	   for(i=0;i< nknots-1;i++){
	      phi[i] = 0.;
	      for(j=0;j< nknots-1;j++)
		 phi[i] = phi[i] + theta[j] * coef2[j][i+2];
	      phi[i] = fabs(phi[i]);
	   }

	/* Invert the information matrix, giving the approximate covariance 
	   matrix for theta                                                           */
	   i= nknots-1;
	   j=NC;
	   xdsifa_(info,&j,&i,kpvt,&infol);
	   xdsidi_(info,&j,&i,kpvt,x1,x2,wk1,&x3);

/* For convenience: symmetarize                                               */
   for(i=1;i< nknots-1;i++)
      for(j=0;j<i;j++)
         info[j][i] = info[i][j];

/* Take linear combinations, to get the standard errors of phi      
   we don't use the tail ones                                                 */
   for(i=1;i< nknots-1;i++){
      se[i] = 0.;
      for(j=0;j< nknots-1;j++)
         for(k=0;k< nknots-1;k++)
            se[i] = se[i] + coef2[j][i+2] * coef2[k][i+2] * info[j][k];
      if(se[i]>0){
      se[i] = sqrt(se[i]);

/* Select for which knot se/phi takes it maximal value */
      if(se[i] > phi[i] * ratmax){
         ratmax = se[i] / phi[i];
         irmax = i;
      }}
   }
   nknots = nknots-1;
   return irmax;
}
/*****************************************************************************/
void fits(xcoef2,xtheta,xctheta,ycoef,xiknots,xnknots)
double xtheta[],xctheta,ycoef[],xcoef2[][NC];
int xiknots[],xnknots;
{
   int i,j;
   for(i=0;i< NC;i++)ycoef[i]=0.;
   ycoef[0]=-log(xctheta);
   for(i=0;i< xnknots;i++){
      ycoef[xiknots[i]+2] = 0.;
      for(j=0;j< xnknots-1;j++)
         ycoef[xiknots[i]+2] = ycoef[xiknots[i]+2] + xtheta[j] * xcoef2[j][i+2];
   }
   for(j=0;j< xnknots-1;j++){
      ycoef[0]=ycoef[0] + xtheta[j] * xcoef2[j][0];
      ycoef[1]=ycoef[1] + xtheta[j] * xcoef2[j][1];
   }
}
