/*******************************************************************************
* Copyright (c) 1993, 1995 by Charles Kooperberg.                              *
* All rights reserved. This function is part of the Logspline Density          *
* Estimation Program. It was written by Charles Kooperberg at the University   *
* of Washington and the University of California at Berkeley between 1988 and  *
* 1991. It is described in the paper: Kooperberg, Charles and Stone, Charles J.*
* `Logspline density estimation for censored data', Journal of Computational   *
* and Graphical Statistics, 1 (1992) 301-328.                                  *
* You are free to use this program, for non-commercial purposes only,          *
* under two conditions:                                                        *
* (1) This note is not to be removed                                           *
* (2) Publications using logspline computations should reference the           *
*  publication  mentioned above.                                               *
* For questions, please email clk@stat.washington.edu                          *
*                       Charles Kooperberg, November 30, 1995                  *
*******************************************************************************/
#include <stdio.h>
#include <math.h>
void pqlsd(coef,knots,bound,ipq,pp,qq,lk,lp)
double coef[],knots[],pp[],bound[],qq[];
int *ipq,*lk,*lp;
{
   void qtop(),ptoq();
   double v1[2],v2[2];
   int ij;
   if((*ipq)==1)
      qtop(coef,knots,bound,pp,qq,*lp,*lk);
   else{
      v2[0]=knots[2];
      ij=1;
      qtop(coef,knots,bound,v1,v2,ij,*lk);
      v2[0]=v2[0];
      for(ij=0;ij<*lp;ij++)pp[ij]=pp[ij]*v2[0];
      ptoq(coef,knots,bound,pp,qq,*lp,*lk,v2[0]);
   }
}
/******************************************************************************/
void ptoq(coef,knots,bound,pp,qq,lp,lk,inp)
double coef[],knots[],bound[],pp[],qq[],inp;
int lp,lk;
{
   double l0,l1,r0,r1,pqexp(),pqnum(),s1,s2,s3,s4,pqexpi(),x1,x2,x3,sj,xj;
   int i,j,vr,vl,k;
   l0 = coef[0];
   l1 = coef[1];
   r0 = l0;
   r1 = l1;
   for(i=0;i<lk;i++){
      r0 = r0 - coef[i+2]*knots[i]*knots[i]*knots[i];
      r1 = r1 + 3.* coef[i+2]*knots[i]*knots[i];
   }
   vr = 4;
   if(bound[2]<0.5)vr=3;
   vl = 2;
   if(bound[0]<0.5)vl=1;
   s4 = inp-pqexp(vr,knots[lk-1],bound[3],r1,r0);
   for(i=0;i<lp;i++){
      j=i;
      x1=pqexpi(vl,bound[1],pp[i],l1,l0);
      if(x1<knots[0])qq[i]=x1;
      else i=lp+100;
   }
   if(i<lp+50)j=lp+100;
   s1 = pqexp(vl,knots[0],bound[1],l1,l0);
   x1 = knots[0];
   s2 = s1;
   x2 = x1;
   k=1;
   s3 = s1 + pqnum(knots[0],knots[1],k,knots,coef);
   x3 = knots[1];
   for(i=j;i<lp;i++){
      if(pp[i]<s4){
         if(pp[i]>=s3){
            do{
               s1 = s3;
               x1 = x3;
               k++;
               s3 = s3 + pqnum(x1,knots[k],k,knots,coef);
               x3 = knots[k];
            }while(pp[i]>s3);
         }
         xj=(knots[k]-knots[k-1])/100.;
         if(k==1||k==lk-1) xj=xj/4.;
         if(pp[i]>=s2){
            s2 = s1;
            x2 = x1;
            do{
               x1 = x2;
               s1 = s2;
               x2 = x1 + xj;
               s2 = s2 + pqnum(x1,x2,k,knots,coef);
            }while(pp[i]>s2);
         }
         sj=(pp[i]-s1)/(s2-s1);
         qq[i]=(1.-sj)*x1+sj*x2;
      }
      else qq[i]=pqexpi(vr,bound[3],(1.-pp[i]),r1,r0);
   }
}
/******************************************************************************/
void qtop(coef,knots,bound,pp,qq,lp,lk)
double coef[],knots[],bound[],pp[],qq[];
int lp,lk;
{
   double l0,l1,r0,r1,s2,pqexp(),pqnum();
   int i,j,k,vr,vl,ko=0;
   l0 = coef[0];
   l1 = coef[1];
   r0 = l0;
   r1 = l1;
   for(i=0;i<lk;i++){
      r0 = r0 - coef[i+2]*knots[i]*knots[i]*knots[i];
      r1 = r1 + 3.* coef[i+2]*knots[i]*knots[i];
   }
   vr = 4;
   if(bound[2]<0.5)vr=3;
   vl = 2;
   if(bound[0]<0.5)vl=1;
   s2 = pqexp(vl,knots[0],bound[1],l1,l0);
   for(i=0;i<lp;i++){
      k=where(qq[i],knots,lk);
      if(k==0)
         pp[i] = pqexp(vl,qq[i],bound[1],l1,l0);
      if(k!=0 && k!= lk){
         if(k==ko){
            s2 = s2 + pqnum(qq[i-1],qq[i],k,knots,coef);
            pp[i] = s2;
         }
         else {
            if(ko!=0)
               s2 = s2 + pqnum(qq[i-1],knots[ko],ko,knots,coef);
            if(k>ko+1)
               for(j=ko+1;j<k;j++)
                  s2 = s2 + pqnum(knots[j-1],knots[j],j,knots,coef);
            s2 = s2 + pqnum(knots[k-1],qq[i],k,knots,coef);
            pp[i] = s2;
         } 
      }
      if(k==lk && ko != lk){
         s2 = s2 + pqnum(qq[i-1],knots[ko],ko,knots,coef);
         if(k>ko+1)
               for(j=ko+1;j<k;j++)
                  s2 = s2 + pqnum(knots[j-1],knots[j],j,knots,coef);
         s2 = s2+pqexp(4,knots[lk-1],qq[i],r1,r0);
         pp[i]=s2;
      }
      if(k==lk && ko == lk){
         s2 = s2+fabs(pqexp(4,qq[i],qq[i-1],r1,r0));
         pp[i]=s2;
      }
      ko = k;
   }
   if( ko != lk){
      s2 = s2 + pqnum(qq[i-1],knots[ko],ko,knots,coef);
      if(lk>ko+1)
         for(j=ko+1;j<lk;j++)
            s2 = s2 + pqnum(knots[j-1],knots[j],j,knots,coef);
      qq[lp-1]=knots[lk-1];
   } 
   s2 = s2+fabs(pqexp(vr,qq[lp-1],bound[3],r1,r0));
   qq[0]=s2;
   for(i=0;i<lp;i++) pp[i]=pp[i]/s2;
}
/******************************************************************************/
double pqexp(version,t1,t2,d,e)
int version;
double t1,t2,d,e;
{
   double f1,f2;
   int i1,i2;
   if(d!=0 || version == 1 || version == 3){
      i1 = 1;
      if(d<0) i1 = -1;
      f1 = log(fabs(1./d)) + d*t1+e;
      if(f1>2000.) f1 = 2000;
      if(version == 1)return i1*exp(f1);
      if(version == 3)return -i1*exp(f1);
      i2 = 1;
      if(d<0) i2 = -1;
      f2 = log(fabs(1./d)) + d*t2+e;
      if(f2>2000.) f2 = 2000;
      if(version==2)return i1*exp(f1)-i2*exp(f2);
      return i2*exp(f2)-i1*exp(f1);
   }
   if(version==4)  return (t2-t1)*exp(e);
   return  (t1-t2)*exp(e);
}
/******************************************************************************/
double pqnum(x1,x2,k,knots,coef)
double x1,x2,knots[],coef[];
int k;
{
   double y[32],w[32],r1,r2,pqdens();
   int i;
   r1 = (x2-x1)/2.;
   r2 = (x2+x1)/2.;
   y[0] = 0.125233408511469 * r1;
   y[1] = 0.367831498998180 * r1;
   y[2] = 0.587317954286617 * r1;
   y[3] = 0.769902674194305 * r1;
   y[4] = 0.904117256370475 * r1;
   y[5] = 0.981560634246719 * r1;
   w[0] = 0.249147045813403 * r1;
   w[1] = 0.233492536538355 * r1;
   w[2] = 0.203167426723066 * r1;
   w[3] = 0.160078328543346 * r1;
   w[4] = 0.106939325995318 * r1;
   w[5] = 0.047175336386512 * r1; 
/* w[ 1]=  0.00178328072169643 * r1; y[1 ]=  0.99930504173577217 * r1;
   w[ 2]=  0.00414703326056247 * r1; y[2 ]=  0.99634011677195533 * r1;
   w[ 3]=  0.00650445796897836 * r1; y[3 ]=  0.99101337147674429 * r1;
   w[ 4]=  0.00884675982636395 * r1; y[4 ]=  0.98333625388462598 * r1;
   w[ 5]=  0.01116813946013113 * r1; y[5 ]=  0.97332682778991098 * r1;
   w[ 6]=  0.01346304789671864 * r1; y[6 ]=  0.96100879965205377 * r1;
   w[ 7]=  0.01572603047602472 * r1; y[7 ]=  0.94641137485840277 * r1;
   w[ 8]=  0.01795171577569734 * r1; y[8 ]=  0.92956917213193957 * r1;
   w[ 9]=  0.02013482315353021 * r1; y[9 ]=  0.91052213707850282 * r1;
   w[10]=  0.02227017380838325 * r1; y[10]=  0.88931544599511414 * r1;
   w[11]=  0.02435270256871087 * r1; y[11]=  0.86599939815409277 * r1;
   w[12]=  0.02637746971505466 * r1; y[12]=  0.84062929625258032 * r1;
   w[13]=  0.02833967261425948 * r1; y[13]=  0.81326531512279754 * r1;
   w[14]=  0.03023465707240248 * r1; y[14]=  0.78397235894334139 * r1;
   w[15]=  0.03205792835485155 * r1; y[15]=  0.75281990726053194 * r1;
   w[16]=  0.03380516183714161 * r1; y[16]=  0.71988185017161088 * r1;
   w[17]=  0.03547221325688239 * r1; y[17]=  0.68523631305423327 * r1;
   w[18]=  0.03705512854024005 * r1; y[18]=  0.64896547125465731 * r1;
   w[19]=  0.03855015317861563 * r1; y[19]=  0.61115535517239328 * r1;
   w[20]=  0.03995374113272034 * r1; y[20]=  0.57189564620263400 * r1;
   w[21]=  0.04126256324262353 * r1; y[21]=  0.53127946401989457 * r1;
   w[22]=  0.04247351512365359 * r1; y[22]=  0.48940314570705296 * r1;
   w[23]=  0.04358372452932345 * r1; y[23]=  0.44636601725346409 * r1;
   w[24]=  0.04459055816375657 * r1; y[24]=  0.40227015796399163 * r1;
   w[25]=  0.04549162792741814 * r1; y[25]=  0.35722015833766813 * r1;
   w[26]=  0.04628479658131442 * r1; y[26]=  0.31132287199021097 * r1;
   w[27]=  0.04696818281621002 * r1; y[27]=  0.26468716220876742 * r1;
   w[28]=  0.04754016571483031 * r1; y[28]=  0.21742364374000708 * r1;
   w[29]=  0.04799938859645831 * r1; y[29]=  0.16964442042399283 * r1;
   w[30]=  0.04834476223480295 * r1; y[30]=  0.12146281929612056 * r1;
   w[31]=  0.04857546744150343 * r1; y[31]=  0.07299312178779904 * r1;
   w[0]=  0.04869095700913972 * r1; y[0]=  0.02435029266342443 * r1; */
   r1 = 0.;
   for(i=0;i<6;i++)
      r1 = r1 + w[i]*(pqdens(r2-y[i],knots,coef,k)+
                      pqdens(r2+y[i],knots,coef,k));
   return r1;
}
/******************************************************************************/
int where(x,knots,lk)
double x,knots[];
int lk;
{
   int i;
   if(x<knots[0])return 0;
   if(x>=knots[lk-1])return lk;
   for(i=1;i<lk-1;i++)
      if(x<knots[i])return i;
   return lk-1;
}
/******************************************************************************/
double pqdens(x,knots,coef,iw)
double x,knots[],coef[];
int iw;
{
   int j;
   double f=coef[0];
   f=f+x*coef[1];
   if(iw>0)
      for(j=0;j<iw;j++)
         f=f+coef[j+2]*pow((x-knots[j]),3.);
   return exp(f);
}
/******************************************************************************/
double pqexpi(version,t,p,d,e)
int version;
double t,p,d,e;
{
   if(d!=0 || version == 1 || version == 3){
   if(version == 1 && p*d < 0)return exp(100000000.);
   if(version == 3 && p*d > 0)return -exp(100000000.);
   if(version == 2 && exp(d*t+e)+p*d < 0)return exp(100000000.);
   if(version == 4 && exp(d*t+e)-p*d < 0)return -exp(100000000.);
   if(version==1)return (log(p*d)-e)/d;
   if(version==2)return (log(exp(d*t+e)+p*d)-e)/d;
   if(version==3)return (log(-p*d)-e)/d;
   return (log(exp(d*t+e)-p*d)-e)/d;
   }
   if(version==2)return t+p/exp(e);
   if(version==4)return t-p/exp(e);
   return t-p/exp(e);
}
/******************************************************************************/
